import re


class Materiale:
    """Overklassen, der ikke benyttes direkte til at skabe objekter (hvilket
    programmet dog ikke ville have problemer med). I stedet bruges den til at
    arve fra af klasserne længere nede i nærværende fil.
    2 af dens funktioner (til_streng() og match_noget) overskrives af objekterne
    dannet på baggrund af underklasserne, mens alle dens øvrige funktioner ikke
    overskrives.
    """
    def __init__(self, idnr=0, titel="N/A", antal=0, antal_udlaan=0,
                 reservationer=0, aarstal=0):
        self.idnr = idnr
        self.titel = titel
        self.antal = antal
        self.antal_udlaan = antal_udlaan
        self.reservationer = reservationer
        self.aarstal = aarstal
        # Denne variabel bliver i underklasserne tildelt en streng med navnet
        # på mediet, dvs. Bog, Film, Lp eller Spil.
        self.medie = "?"

    # Sørger for en fin og ensartet formatering, inden der udskrives
    # til tekstfeltet. Fx betyder %3d, at tallet udskrives højrestillet i
    # 3 karakterfelter.
    def til_streng(self):
        info = "ID:%3d   Titel: %-25s   Medie: %-4s   År: %4d   Antal:%2d" \
               "   Udlånt:%2d   Res.:%2d"\
               % (self.idnr, self.titel, self.medie, self.aarstal,
                  self.antal, self.antal_udlaan, self.reservationer)
        return info

    # Handler om almindeligt udlån, dvs. udlån, som ikke sker på baggrund af
    # en reservation.
    # Returnerer "True", hvis der er flere bøger hjemme, end der er reserveret,
    def udlaan_muligt(self):
        if self.antal - (self.antal_udlaan + self.reservationer) > 0:
            return True
        else:
            return False

    def udlaan(self):
        self.antal_udlaan += 1

    # Er med til at sørge for, at antal udlån ikke lige pludselig bliver
    # negativ. Der er andre måder at sørge for dette på, men handlingen er jo
    # simpelthen umulig i den virkelige verden, hvorfor der skal gøres
    # opmærksom derpå, ligesom handlingen skal stoppes i udgangspunktet.
    def aflever_muligt(self):
        if self.antal_udlaan > 0:
            return True
        else:
            return False

    def aflever(self):
        self.antal_udlaan -= 1

    def reserver(self):
        self.reservationer += 1

    # Returnerer "True", hvis der burde blive indkøbt flere eksemplarer af
    # materialet.
    # (Jeg har dog ikke prioriteret at indbygge indkøb af ekstra eksemplarer i
    # programmet.)
    def for_lille_udbud(self):
        if self.reservationer > 5 and self.reservationer % 3 == 0 and \
                self.reservationer > 2 * (self.antal - self.antal_udlaan):
            return True
        else:
            return False

    def fjern_reservation(self):
        self.reservationer -= 1

    # Udlån på baggrund af reservation kan kun ske, hvis der er mindst en
    # reservation og hvis der er bøger hjemme.
    def laan_reservation_muligt(self):
        if self.reservationer > 0 and self.antal > self.antal_udlaan:
            return True
        else:
            return False

    # Modulet "re", som står for "regular expressions", benyttes her og i den
    # næste funktion.
    # Fra "re" benyttes "search"-funktionen, der kigger efter et match mellem
    # soege_tekst og hvor som helst i den streng, den kaldes på.
    # IGNORECASE sørger for, at et bestemt bogstav skrevet med småt anses for
    # at matche det samme bogstav skrevet med stort og omvendt.
    def match_titel(self, soege_tekst):
        if re.search(soege_tekst, self.titel, re.IGNORECASE):
            return True
        else:
            return False

    def match_aarstal(self, soege_tekst):
        if re.search(soege_tekst, str(self.aarstal)):
            return True
        else:
            return False

    # Tjekker sammenfald mellem søgetekst
    # og et materiales titel og/eller aarstal.
    def match_noget(self, soege_tekst):
        if self.match_titel(soege_tekst) or self.match_aarstal(soege_tekst):
            return True
        else:
            return False


class Bog(Materiale):
    """En af fire underklasser, der arver fra Materiale."""

    def __init__(self, idnr=0, titel="N/A", antal=0, antal_udlaan=0,
                 reservationer=0, aarstal=0, forfatter="N/A", antal_sider=0):
        # Her i underklassens kunstruktør kaldes overklassen "Materiale"s
        # konstruktør med de værdier, som alle objekter af klassen Materiale
        # har.
        super(Bog, self).__init__(idnr, titel, antal, antal_udlaan,
                                  reservationer, aarstal)
        # Her har vi så de nye attributter, som er særlige for denne klasse.
        # Derudover sættes medie til "Bog", da alle objekter dannet ud fra
        # denne klasse nødvendigvis er bøger.
        self.forfatter = forfatter
        self.antal_sider = antal_sider
        self.medie = "Bog"

    def til_streng(self):
        info = "ID:%3d   Titel: %-25s   Medie: %-4s   År: %4d   Antal:%2d" \
               "   Udlånt:%2d   Res.:%2d   Forfatter: %-25s   Sider: %d" \
               % (self.idnr, self.titel, self.medie, self.aarstal,
                  self.antal, self.antal_udlaan, self.reservationer,
                  self.forfatter, self.antal_sider)
        return info

    # Tjekker sammenfald mellem søgetekst
    # og en bogs titel, aarstal og/eller forfatter.
    def match_noget(self, soege_tekst):
        if self.match_titel(soege_tekst) or self.match_aarstal(soege_tekst) \
                or re.search(soege_tekst, self.forfatter, re.IGNORECASE):
            return True
        else:
            return False


class Film(Materiale):

    def __init__(self, idnr=0, titel="N/A", antal=0, antal_udlaan=0,
                 reservationer=0, aarstal=0, instruktoer="N/A", varighed=0):
        super(Film, self).__init__(idnr, titel, antal, antal_udlaan,
                                   reservationer, aarstal)
        self.instruktoer = instruktoer
        self.varighed = varighed
        self.medie = "Film"

    def til_streng(self):
        info = "ID:%3d   Titel: %-25s   Medie: %-4s   År: %4d   Antal:%2d" \
               "   Udlånt:%2d   Res.:%2d   Instruktør: %-24s   " \
               "Varighed: %d min." \
               % (self.idnr, self.titel, self.medie, self.aarstal,
                  self.antal, self.antal_udlaan, self.reservationer,
                  self.instruktoer, self.varighed)
        return info

    # Tjekker sammenfald mellem søgetekst
    # og en films titel, aarstal og/eller instruktør.
    def match_noget(self, soege_tekst):
        if self.match_titel(soege_tekst) or self.match_aarstal(soege_tekst) \
                or re.search(soege_tekst, self.instruktoer, re.IGNORECASE):
            return True
        else:
            return False


class Lp(Materiale):

    def __init__(self, idnr=0, titel="N/A", antal=0, antal_udlaan=0,
                 reservationer=0, aarstal=0, kunstner="N/A", varighed=0):
        super(Lp, self).__init__(idnr, titel, antal, antal_udlaan,
                                 reservationer, aarstal)
        self.kunstner = kunstner
        self.varighed = varighed
        self.medie = "Lp"

    def til_streng(self):
        info = "ID:%3d   Titel: %-25s   Medie: %-4s   År: %4d   Antal:%2d" \
               "   Udlånt:%2d   Res.:%2d   Kunstner: %-26s   Varighed: %d min."\
               % (self.idnr, self.titel, self.medie, self.aarstal,
                  self.antal, self.antal_udlaan, self.reservationer,
                  self.kunstner, self.varighed)
        return info

    # Tjekker sammenfald mellem søgetekst
    # og en Lps titel, aarstal og/eller kunstner.
    def match_noget(self, soege_tekst):
        if self.match_titel(soege_tekst) or self.match_aarstal(soege_tekst) \
                or re.search(soege_tekst, self.kunstner, re.IGNORECASE):
            return True
        else:
            return False


class Spil(Materiale):

    def __init__(self, idnr=0, titel="N/A", antal=0, antal_udlaan=0,
                 reservationer=0, aarstal=0, udgiver="N/A", konsol="N/A"):
        super(Spil, self).__init__(idnr, titel, antal, antal_udlaan,
                                   reservationer, aarstal)
        self.udgiver = udgiver
        self.konsol = konsol
        self.medie = "Spil"

    def til_streng(self):
        info = "ID:%3d   Titel: %-25s   Medie: %-4s   År: %4d   Antal:%2d" \
               "   Udlånt:%2d   Res.:%2d   Udgiver: %-27s   Konsol: %s" \
               % (self.idnr, self.titel, self.medie, self.aarstal,
                  self.antal, self.antal_udlaan, self.reservationer,
                  self.udgiver, self.konsol)
        return info

    # Tjekker sammenfald mellem søgetekst
    # og et spils titel, aarstal, udgiver og/eller konsol.
    def match_noget(self, soege_tekst):
        if self.match_titel(soege_tekst) or self.match_aarstal(soege_tekst) \
                or re.search(soege_tekst, self.udgiver, re.IGNORECASE)\
                or re.search(soege_tekst, self.konsol, re.IGNORECASE):
            return True
        else:
            return False
